from abc import ABC, abstractmethod
from typing import List
import re


class Tokenizer(ABC):
    @abstractmethod
    def tokenize(self, text: str) -> List[str]:
        pass


# gensim simple_tokenize pattern
PAT_ALPHABETIC = re.compile(r"(((?![\d])\w)+)", re.UNICODE)


class SimpleTokenizer(Tokenizer):
    def tokenize(self, text: str) -> List[str]:
        tokens = [match.group() for match in PAT_ALPHABETIC.finditer(text)]
        return tokens


class NGramTokenizer(Tokenizer):
    def __init__(self, min_gram: int, max_gram: int):
        super().__init__()
        self.min_gram = min_gram
        self.max_gram = max_gram

    def _convert_text_to_ngrams(self, text: str):
        if len(text) <= self.min_gram:
            return [text]

        ngrams = []
        for i in range(len(text)):
            # max_gram + 1 because range is not inclusive to last number
            for j in range(self.min_gram, self.max_gram + 1):
                if i + j > len(text):
                    break
                ngrams.append(text[i : i + j])

        return ngrams

    def tokenize(self, text: str) -> List[str]:
        return self._convert_text_to_ngrams(text)



from src.textsearchpy.tokenizers import SimpleTokenizer, NGramTokenizer


def test_simple_tokenizer():
    text = "Iteratively yield tokens as unicode strings, removing accent marks and optionally lowercasing the unidoce string by assigning True to one of the parameters, lowercase, to_lower, or lower."
    tokenizer = SimpleTokenizer()
    tokens = tokenizer.tokenize(text)
    assert tokens == [
        "Iteratively",
        "yield",
        "tokens",
        "as",
        "unicode",
        "strings",
        "removing",
        "accent",
        "marks",
        "and",
        "optionally",
        "lowercasing",
        "the",
        "unidoce",
        "string",
        "by",
        "assigning",
        "True",
        "to",
        "one",
        "of",
        "the",
        "parameters",
        "lowercase",
        "to_lower",
        "or",
        "lower",
    ]

    text = "Version 4.0 was released on October 12, 2012."
    tokens = tokenizer.tokenize(text)
    assert tokens == ["Version", "was", "released", "on", "October"]


def test_ngram_tokenizer():
    tokenizer = NGramTokenizer(min_gram=5, max_gram=5)
    text = "Quick Fox"
    tokens = tokenizer.tokenize(text)

    assert tokens == ["Quick", "uick ", "ick F", "ck Fo", "k Fox"]

    tokenizer = NGramTokenizer(min_gram=5, max_gram=6)
    tokens = tokenizer.tokenize(text)
    assert tokens == [
        "Quick",
        "Quick ",
        "uick ",
        "uick F",
        "ick F",
        "ick Fo",
        "ck Fo",
        "ck Fox",
        "k Fox",
    ]



from benchmark_utils import create_index_from_data, print_memory_usage, evaluate_queries
from textsearchpy.query import TermQuery, BooleanQuery, BooleanClause, PhraseQuery
from os import listdir
from os.path import isfile, join
from pathlib import Path

# https://www.nltk.org/nltk_data/ The Reuters-21578 benchmark corpus
REUTERS_DATA_PATH = "/Users/kimili/Downloads/reuters/total"


def load_corpus():
    data = []
    total_size = 0
    for f in listdir(REUTERS_DATA_PATH):
        path = join(REUTERS_DATA_PATH, f)
        if isfile(path):
            with open(join(REUTERS_DATA_PATH, f), "r", encoding="latin-1") as file:
                data.append(file.read())

            total_size += Path(path).stat().st_size

    print(f"Raw data size: {total_size * 0.000001} MB")
    return data


def run():
    data = load_corpus()
    print(f"Reuters data loaded with {len(data)} documents")

    print_memory_usage()

    index = create_index_from_data(data)

    print_memory_usage()

    evaluate_queries(
        index=index,
        queries=[
            TermQuery(term="payout"),
            TermQuery(term="income"),
            TermQuery(term="inventory"),
        ],
    )

    evaluate_queries(
        index=index,
        queries=[
            BooleanQuery(
                clauses=[
                    BooleanClause(query=TermQuery(term="payout"), clause="MUST"),
                    BooleanClause(query=TermQuery(term="income"), clause="MUST"),
                ]
            )
        ],
    )

    evaluate_queries(
        index=index,
        queries=[
            PhraseQuery(terms=["management", "acquisition"], distance=2, ordered=False)
        ],
    )


if __name__ == "__main__":
    run()



from abc import ABC
from enum import Enum
from typing import List
from pydantic import BaseModel


class Query(BaseModel, ABC):
    pass


# no plans for fields for now
class TermQuery(Query):
    term: str


class Clause(str, Enum):
    SHOULD = "SHOULD"
    MUST = "MUST"
    MUST_NOT = "MUST_NOT"


class BooleanClause(BaseModel):
    query: Query
    clause: Clause


class BooleanQuery(Query):
    clauses: List[BooleanClause]


class PhraseQuery(Query):
    terms: List[str]
    distance: int
    ordered: bool = False


class QueryParseError(Exception):
    pass


RESERVED_KEY_CHAR = set(["(", ")", '"'])


def tokenize_query(query: str) -> List[str]:
    tokens = []
    buffer = []

    for char in query:
        if char in RESERVED_KEY_CHAR:
            if len(buffer) > 0:
                tokens.append("".join(buffer))
                buffer.clear()
            tokens.append(char)
        elif char.isspace():
            if len(buffer) > 0:
                tokens.append("".join(buffer))
                buffer.clear()
        else:
            buffer.append(char)
    if len(buffer) > 0:
        tokens.append("".join(buffer))

    return tokens


def parse_query(query: str) -> Query:
    tokens = tokenize_query(query)
    return _parse_group_q(tokens)


def _parse_group_q(tokens: List[str]) -> Query:
    left = _parse_base_q(tokens)
    if len(tokens) == 0:
        return left

    query_clauses = []
    query_clauses.append(BooleanClause(query=left, clause=Clause.SHOULD))
    while tokens and tokens[0] != ")":
        op_type = None
        if tokens[0] in ["AND", "OR", "NOT"]:
            op_type = tokens.pop(0)
        else:
            # implied op is OR i.e. query = "word1 word2" = "word1 OR word2"
            op_type = "OR"

        right = _parse_base_q(tokens)

        if op_type == "OR":
            query_clauses.append(BooleanClause(query=right, clause=Clause.SHOULD))
        elif op_type == "NOT":
            query_clauses.append(BooleanClause(query=right, clause=Clause.MUST_NOT))
        elif op_type == "AND":
            # in cases of "word1 AND word2" word1 should also become MUST conditions
            if query_clauses[-1].clause == Clause.SHOULD:
                query_clauses[-1] = BooleanClause(
                    query=query_clauses[-1].query, clause=Clause.MUST
                )
            query_clauses.append(BooleanClause(query=right, clause=Clause.MUST))

    return BooleanQuery(clauses=query_clauses)


def _parse_base_q(tokens: List[str]):
    token = tokens[0]

    # group query
    if token == "(":
        tokens.pop(0)
        q = _parse_group_q(tokens)
        if len(tokens) == 0 or tokens[0] != ")":
            raise QueryParseError("Discovered Un-ending Bracket")
        tokens.pop(0)
        return q

    # phrase query
    elif token == '"':
        tokens.pop(0)
        if len(tokens) < 3:
            raise QueryParseError("Malformed Phrase Query Discovered")

        terms = []
        while tokens and tokens[0] != '"':
            terms.append(tokens.pop(0))

        if tokens[0] != '"':
            raise QueryParseError(
                "Phrase Query detected with unterminated double quote"
            )

        tokens.pop(0)
        distance = None
        if tokens and tokens[0].startswith("~"):
            distance_str = tokens.pop(0)[1:]
            try:
                distance = int(distance_str)
            except ValueError:
                raise QueryParseError(
                    f"Failed to parse phrase query edit distance: found {distance}"
                )

        distance = distance if distance else 0

        return PhraseQuery(terms=terms, distance=distance)

    # term query
    else:
        term = tokens.pop(0)
        q = TermQuery(term=term)
        return q



import json
import pytest
from src.textsearchpy.index import Document, Index, IndexingError
from src.textsearchpy.query import BooleanClause, BooleanQuery, PhraseQuery, TermQuery
from src.textsearchpy.normalizers import StopwordsNormalizer
import os


def test_append_doc():
    index = Index()
    assert len(index.documents) == 0
    assert len(index.inverted_index) == 0

    doc1 = Document(
        text="repository contains all project files, including the revision history"
    )
    index.append([doc1])

    assert len(index.documents) == 1
    assert len(index.inverted_index) == 9

    doc2 = Document(text="repository repeats words words words")
    doc3 = Document(text="words and words and words")
    index.append([doc2, doc3])
    assert len(index.documents) == 3
    assert len(index.inverted_index) == 12


def test_append_doc_with_id():
    index = Index()
    doc1 = Document(text="i like cake", id="1")
    doc2 = Document(text="you like cookie", id="2")
    index.append([doc1, doc2])

    assert len(index) == 2
    assert index.documents["1"].text == "i like cake"
    assert index.documents["2"].text == "you like cookie"

    with pytest.raises(IndexingError):
        dupe_id_doc = Document(text="another doc", id="1")
        index.append([dupe_id_doc])


def test_append_doc_mixed_type():
    index = Index()

    doc1 = Document(text="abcd")
    doc2 = "qwer"

    index.append([doc1, doc2])
    assert len(index.documents) == 2


def test_positional_index_append():
    index = Index()

    doc1 = Document(text="this book has a lot of words for a book")
    doc2 = Document(text="can you give this book away for me")

    index.append([doc1, doc2])

    assert len(index.positional_index.keys()) == 13
    assert len(index.positional_index["book"].keys()) == 2
    assert len(index.positional_index["a"].keys()) == 1
    assert len(index.positional_index["away"].keys()) == 1

    # get the doc_id
    doc1_id = list(index.positional_index["a"].keys())[0]
    doc2_id = list(index.positional_index["away"].keys())[0]

    assert index.documents[doc1_id].text == doc1.text
    assert index.documents[doc2_id].text == doc2.text

    assert index.positional_index["book"][doc1_id] == [1, 9]
    assert index.positional_index["book"][doc2_id] == [4]


def test_search():
    index = Index()

    doc1 = Document(text="i like cake")
    doc2 = Document(text="you like cookie")
    doc3 = Document(text="we like cake")

    index.append([doc1, doc2, doc3])

    docs = index.search("like")
    assert len(docs) == 3

    docs = index.search("you")
    assert len(docs) == 1
    assert docs[0].text == "you like cookie"

    docs = index.search("cake")
    assert len(docs) == 2

    docs = index.search("what")
    assert len(docs) == 0


def test_term_search():
    index = Index()

    doc1 = Document(text="i like cake")
    doc2 = Document(text="you like cookie")
    doc3 = Document(text="we like cake")

    index.append([doc1, doc2, doc3])

    q = TermQuery(term="cookie")
    docs = index.search(q)
    assert len(docs) == 1

    q = TermQuery(term="cake")
    docs = index.search(q)
    assert len(docs) == 2


def test_boolean_search():
    index = Index()

    doc1 = Document(text="i like cake")
    doc2 = Document(text="you like cookie")
    doc3 = Document(text="we like cake")
    doc4 = Document(text="we should have a tea party")

    index.append([doc1, doc2, doc3, doc4])

    q1 = TermQuery(term="cookie")
    q2 = TermQuery(term="cake")
    q = BooleanQuery(
        clauses=[
            BooleanClause(query=q1, clause="SHOULD"),
            BooleanClause(query=q2, clause="SHOULD"),
        ]
    )
    docs = index.search(q)
    assert len(docs) == 3

    q1 = TermQuery(term="like")
    q2 = TermQuery(term="we")
    q = BooleanQuery(
        clauses=[
            BooleanClause(query=q1, clause="MUST"),
            BooleanClause(query=q2, clause="MUST"),
        ]
    )
    docs = index.search(q)
    assert len(docs) == 1

    q1 = TermQuery(term="cake")
    q2 = TermQuery(term="like")
    q = BooleanQuery(
        clauses=[
            BooleanClause(query=q1, clause="MUST_NOT"),
            BooleanClause(query=q2, clause="SHOULD"),
        ]
    )
    docs = index.search(q)
    assert len(docs) == 1

    q1 = TermQuery(term="cake")
    q2 = TermQuery(term="cookie")
    q = BooleanQuery(
        clauses=[
            BooleanClause(query=q1, clause="MUST"),
            BooleanClause(query=q2, clause="SHOULD"),
        ]
    )
    docs = index.search(q)
    assert len(docs) == 2


def test_phrase_query():
    index = Index()

    doc1 = Document(text="i like cake, but do we like this specific cake")
    doc2 = Document(text="you like cookie")
    doc3 = Document(text="we like cake")
    doc4 = Document(text="we should have a tea party")
    index.append([doc1, doc2, doc3, doc4])

    q = PhraseQuery(terms=["like", "cake"], distance=0)
    docs = index.search(q)
    assert len(docs) == 2

    q = PhraseQuery(terms=["we", "cake"], distance=1)
    docs = index.search(q)
    assert len(docs) == 1

    # currently algo is not order sensitive
    q = PhraseQuery(terms=["cake", "like"], distance=0)
    docs = index.search(q)
    assert len(docs) == 2

    q = PhraseQuery(terms=["we", "cake"], distance=2)
    docs = index.search(q)
    assert len(docs) == 2

    q = PhraseQuery(terms=["we", "cake"], distance=0)
    docs = index.search(q)
    assert len(docs) == 0

    q = PhraseQuery(terms=["we", "cookie"], distance=0)
    docs = index.search(q)
    assert len(docs) == 0


def test_phrase_query_with_same_word():
    index = Index()
    doc = Document(text="you like cookie")
    index.append([doc])
    # should not match because there is not two like tokens
    q = PhraseQuery(terms=["like", "like"], distance=0)
    docs = index.search(q)
    assert len(docs) == 0

    doc2 = Document(text="you like like cookie")
    index.append([doc2])
    docs = index.search(q)
    assert len(docs) == 1


def test_phrase_query_ordered():
    index = Index()
    doc = Document(text="you like cookie")
    index.append([doc])

    q = PhraseQuery(terms=["cookie", "you"], distance=1, ordered=False)
    docs = index.search(q)
    assert len(docs) == 1

    q = PhraseQuery(terms=["cookie", "you"], distance=1, ordered=True)
    docs = index.search(q)
    assert len(docs) == 0

    q = PhraseQuery(terms=["you", "cookie"], distance=1, ordered=True)
    docs = index.search(q)
    assert len(docs) == 1


def test_multi_term_phrase_query():
    index = Index()

    doc1 = Document(text="i like cake, but do we like this specific cake")
    doc2 = Document(text="you like cookie")
    doc3 = Document(text="we like cake")
    doc4 = Document(text="we should have a tea party")
    doc5 = Document(text="cake like i")
    index.append([doc1, doc2, doc3, doc4, doc5])

    q = PhraseQuery(terms=["i", "like", "cake"], distance=0)
    docs = index.search(q)
    assert len(docs) == 2

    q = PhraseQuery(terms=["like", "i", "cake"], distance=0)
    docs = index.search(q)
    assert len(docs) == 2

    q = PhraseQuery(terms=["we", "like", "cake"], distance=0)
    docs = index.search(q)
    assert len(docs) == 1

    q = PhraseQuery(terms=["we", "like", "cake"], distance=1)
    docs = index.search(q)
    assert len(docs) == 1

    q = PhraseQuery(terms=["we", "like", "cake"], distance=2)
    docs = index.search(q)
    assert len(docs) == 2


def test_multi_term_phrase_ordered_query():
    index = Index()

    doc1 = Document(text="i like cake, but do we like this specific cake")
    doc2 = Document(text="you like cookie")
    doc3 = Document(text="we like cake")
    doc4 = Document(text="we should have a tea party")
    doc5 = Document(text="cake like i")
    index.append([doc1, doc2, doc3, doc4, doc5])

    q = PhraseQuery(terms=["like", "i", "cake"], distance=0, ordered=True)
    docs = index.search(q)
    assert len(docs) == 0

    q = PhraseQuery(terms=["i", "like", "cake"], distance=0, ordered=True)
    docs = index.search(q)
    assert len(docs) == 1

    q = PhraseQuery(terms=["we", "like", "cake"], distance=0, ordered=True)
    docs = index.search(q)
    assert len(docs) == 1

    q = PhraseQuery(terms=["we", "like", "cake"], distance=1, ordered=True)
    docs = index.search(q)
    assert len(docs) == 1

    q = PhraseQuery(terms=["we", "like", "cake"], distance=2, ordered=True)
    docs = index.search(q)
    assert len(docs) == 2


def test_string_query():
    index = Index()
    doc1 = Document(text="i like cake, but do we like this specific cake")
    doc2 = Document(text="you like cookie")
    doc3 = Document(text="we like cake")
    doc4 = Document(text="we should have a tea party")
    index.append([doc1, doc2, doc3, doc4])

    q = "tea"
    docs = index.search(q)
    assert len(docs) == 1

    q = "cake cookie"
    docs = index.search(q)
    assert len(docs) == 3

    q = "cake AND specific"
    docs = index.search(q)
    assert len(docs) == 1

    q = "cake NOT like"
    docs = index.search(q)
    assert len(docs) == 0


def test_index_length():
    index = Index()
    assert len(index) == 0

    doc1 = Document(text="you like cookie")
    doc2 = Document(text="we like cake")
    index.append([doc1, doc2])

    assert len(index) == 2


def test_index_delete():
    index = Index()
    doc1 = Document(text="i like cake, but do we like this specific cake", id="1")
    doc2 = Document(text="you like cookie", id="2")
    doc3 = Document(text="we like cake", id="3")
    doc4 = Document(text="we should have a tea party", id="4")
    index.append([doc1, doc2, doc3, doc4])

    index.delete(ids=["1", "2", "3"])

    assert len(index) == 1
    assert index.inverted_index["we"] == ["4"]
    assert index.positional_index["we"] == {"4": [0]}

    assert len(index.search("cake")) == 0
    assert len(index.search("tea")) == 1


def test_query_with_filtered_tokens():
    index = Index(token_normalizers=[StopwordsNormalizer()])

    doc1 = Document(text="i like cake")
    index.append([doc1])

    # searching a term filtered by StopwordsNormalizer
    q = TermQuery(term="i")
    docs = index.search(q)
    assert len(docs) == 0


def test_index_save_load(tmp_path):
    index = Index()
    doc1 = Document(text="you like cookie")
    doc2 = Document(text="we like cake")
    index.append([doc1, doc2])

    save_path = str(tmp_path / "test_save")
    index.save(path=save_path)

    index_file = os.path.join(save_path, "index.json")
    doc_file = os.path.join(save_path, "docs.jsonl")
    assert os.path.exists(index_file)
    assert os.path.exists(doc_file)

    with open(index_file, "r") as f:
        saved_index_file = json.load(f)

    assert saved_index_file["token_normalizers"] == ["LowerCaseNormalizer"]
    assert saved_index_file["tokenizer"] == "SimpleTokenizer"
    assert len(saved_index_file["inverted_index"]) == 5
    assert len(saved_index_file["positional_index"]) == 5

    saved_docs = []
    with open(doc_file, "r") as f:
        for line in f:
            saved_docs.append(json.loads(line))

    assert saved_docs[0]["text"] == "you like cookie"
    assert saved_docs[1]["text"] == "we like cake"

    new_index = Index()
    new_index.load_from_file(save_path)
    assert len(new_index.search("you")) == 1
    assert len(new_index.search("like")) == 2
    assert len(new_index.documents) == 2



from src.textsearchpy.query import BooleanQuery, PhraseQuery, TermQuery, parse_query


def test_parse_term_query():
    query = "word"
    q = parse_query(query)
    assert isinstance(q, TermQuery)
    assert q.term == "word"


def test_basic_boolean_query():
    query = "word search"
    q = parse_query(query)
    assert isinstance(q, BooleanQuery)
    assert len(q.clauses) == 2
    assert isinstance(q.clauses[0].query, TermQuery)
    assert q.clauses[0].query.term == "word"
    assert q.clauses[0].clause == "SHOULD"
    assert isinstance(q.clauses[1].query, TermQuery)
    assert q.clauses[1].query.term == "search"
    assert q.clauses[1].clause == "SHOULD"

    query = "word AND search"
    q = parse_query(query)
    assert isinstance(q, BooleanQuery)
    assert len(q.clauses) == 2
    assert isinstance(q.clauses[0].query, TermQuery)
    assert q.clauses[0].query.term == "word"
    assert q.clauses[0].clause == "MUST"
    assert isinstance(q.clauses[1].query, TermQuery)
    assert q.clauses[1].query.term == "search"
    assert q.clauses[1].clause == "MUST"

    query = "word NOT search"
    q = parse_query(query)
    assert isinstance(q, BooleanQuery)
    assert len(q.clauses) == 2
    assert isinstance(q.clauses[0].query, TermQuery)
    assert q.clauses[0].query.term == "word"
    assert q.clauses[0].clause == "SHOULD"
    assert isinstance(q.clauses[1].query, TermQuery)
    assert q.clauses[1].query.term == "search"
    assert q.clauses[1].clause == "MUST_NOT"


def test_compound_boolean_query():
    query = "word AND search NOT found"
    q = parse_query(query)
    assert isinstance(q, BooleanQuery)
    assert len(q.clauses) == 3
    assert isinstance(q.clauses[0].query, TermQuery)
    assert q.clauses[0].query.term == "word"
    assert q.clauses[0].clause == "MUST"
    assert isinstance(q.clauses[1].query, TermQuery)
    assert q.clauses[1].query.term == "search"
    assert q.clauses[1].clause == "MUST"
    assert isinstance(q.clauses[2].query, TermQuery)
    assert q.clauses[2].query.term == "found"
    assert q.clauses[2].clause == "MUST_NOT"

    query = "word OR search NOT found"
    q = parse_query(query)
    assert isinstance(q, BooleanQuery)
    assert len(q.clauses) == 3
    assert isinstance(q.clauses[0].query, TermQuery)
    assert q.clauses[0].query.term == "word"
    assert q.clauses[0].clause == "SHOULD"
    assert isinstance(q.clauses[1].query, TermQuery)
    assert q.clauses[1].query.term == "search"
    assert q.clauses[1].clause == "SHOULD"
    assert isinstance(q.clauses[2].query, TermQuery)
    assert q.clauses[2].query.term == "found"
    assert q.clauses[2].clause == "MUST_NOT"


def test_basic_phrase_query():
    query = '"word search"'
    q = parse_query(query)
    assert isinstance(q, PhraseQuery)
    assert q.terms == ["word", "search"]
    assert q.distance == 0

    query = '"word search"~5'
    q = parse_query(query)
    assert isinstance(q, PhraseQuery)
    assert q.terms == ["word", "search"]
    assert q.distance == 5


def test_multi_term_phrase_query():
    query = '"multi word search"~3'
    q = parse_query(query)
    assert isinstance(q, PhraseQuery)
    assert q.terms == ["multi", "word", "search"]
    assert q.distance == 3


def test_basic_group_query():
    query = "(group word) AND search"
    q = parse_query(query)
    assert isinstance(q, BooleanQuery)
    assert isinstance(q.clauses[0].query, BooleanQuery)
    assert q.clauses[0].clause == "MUST"
    assert isinstance(q.clauses[1].query, TermQuery)
    assert q.clauses[1].clause == "MUST"

    sub_q = q.clauses[0].query
    assert isinstance(sub_q.clauses[1].query, TermQuery)
    assert sub_q.clauses[0].clause == "SHOULD"
    assert isinstance(sub_q.clauses[1].query, TermQuery)
    assert sub_q.clauses[1].clause == "SHOULD"



from src.textsearchpy.normalizers import (
    LowerCaseNormalizer,
    StopwordsNormalizer,
)


def test_lowercase_normalizer():
    normalizer = LowerCaseNormalizer()

    tokens = ["Hi", "JOHN", "what"]
    tokens = normalizer.normalize(tokens)

    assert tokens == ["hi", "john", "what"]


def test_stopwords_normalizer():
    normalizer = StopwordsNormalizer()

    tokens = ["what", "is", "your", "name"]
    t_tokens = normalizer.normalize(tokens)
    assert t_tokens == ["name"]

    tokens = ["rare", "search"]
    t_tokens = normalizer.normalize(tokens)
    assert tokens == t_tokens


def test_stopwrods_normalizer_custom_words():
    stopwords = ["mock"]
    normalizer = StopwordsNormalizer(stopwords=stopwords)

    tokens = ["this", "is", "mock", "testing"]
    t_tokens = normalizer.normalize(tokens)
    assert t_tokens == ["this", "is", "testing"]



from benchmark_utils import create_index_from_data, print_memory_usage, evaluate_queries
from textsearchpy.query import TermQuery
from os import listdir
from os.path import isfile, join, isdir
import argparse
from pathlib import Path

# sample of project gutenberg books
DATA_PATH = "/Users/kimili/Downloads/gutenberg_1000"


def load_corpus(n):
    data = []
    total_size = 0
    for dir in listdir(DATA_PATH):
        if isdir(join(DATA_PATH, dir)):
            for f in listdir(join(DATA_PATH, dir)):
                path = join(DATA_PATH, dir, f)
                if isfile(path):
                    with open(path, "r") as file:
                        data.append(file.read())

                    total_size += Path(path).stat().st_size

                    if len(data) == n:
                        print(f"Raw data size: {total_size * 0.000001} MB")
                        return data

    print(f"Raw data size: {total_size * 0.000001} MB")
    return data


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument("-n", type=int)

    args = parser.parse_args()

    data = load_corpus(args.n)
    print(f"Gutenberg data loaded with {len(data)} documents")

    print_memory_usage()

    index = create_index_from_data(data)

    print_memory_usage()

    evaluate_queries(
        index=index,
        queries=[
            TermQuery(term="hearts"),
            TermQuery(term="child"),
            TermQuery(term="software"),
        ],
    )


if __name__ == "__main__":
    run()



import json
from pathlib import Path
from typing import Dict, List, Optional, Union
from pydantic import BaseModel
import uuid
import os

from .tokenizers import SimpleTokenizer, Tokenizer
from .normalizers import TokenNormalizer, LowerCaseNormalizer
from .query import (
    BooleanQuery,
    Clause,
    PhraseQuery,
    Query,
    TermQuery,
    parse_query,
)


class IndexSearchError(Exception):
    pass


class IndexingError(Exception):
    pass


class Document(BaseModel):
    text: str
    # metadata
    id: Optional[str] = None


class Index:
    def __init__(
        self,
        token_normalizers: List[TokenNormalizer] = [LowerCaseNormalizer()],
        tokenizer: Tokenizer = SimpleTokenizer(),
    ):
        self.token_normalizers: List[TokenNormalizer] = token_normalizers
        self.tokenizer: Tokenizer = tokenizer

        # {token: doc_id}
        self.inverted_index: Dict[str, List[str]] = {}
        self.documents: Dict[str, Document] = {}
        # {token: {doc_id: [token_index]}}
        self.positional_index: Dict[str, Dict[str, List[int]]] = {}

    def __len__(self):
        return len(self.documents)

    def _add_to_index(self, doc: Document, tokens: List[str]):
        if doc.id is None:
            raise ValueError("Document ID cannot be None")

        self.documents[doc.id] = doc

        if tokens:
            for tok_i, tok in enumerate(tokens):
                # add to inverted index, first check if doc id already has been recorded in positional index to avoid dupe ids
                # TODO think about if inverted_index is needed vs using positional index only
                if (
                    tok not in self.positional_index
                    or doc.id not in self.positional_index[tok]
                ):
                    if tok in self.inverted_index:
                        self.inverted_index[tok].append(doc.id)
                    else:
                        self.inverted_index[tok] = [doc.id]

                # add to positional index
                if tok not in self.positional_index:
                    self.positional_index[tok] = {}

                if doc.id in self.positional_index[tok]:
                    self.positional_index[tok][doc.id].append(tok_i)
                else:
                    self.positional_index[tok][doc.id] = [tok_i]

    def _normalize_tokens(self, tokens: List[str]):
        if not self.token_normalizers:
            return tokens

        for normalizer in self.token_normalizers:
            tokens = normalizer.normalize(tokens)

        return tokens

    def text_to_index_tokens(self, text: str) -> List[str]:
        tokens = self.tokenizer.tokenize(text)
        tokens = self._normalize_tokens(tokens)
        return tokens

    def append(self, docs: List[Union[str, Document]]):
        for doc in docs:
            if isinstance(doc, str):
                doc = Document(text=doc)

            tokens = self.text_to_index_tokens(doc.text)
            if doc.id is not None:
                if doc.id in self.documents:
                    raise IndexingError(
                        f"Attempting to add a Document with ID: {doc.id} already exists in index"
                    )
            else:
                doc_id = uuid.uuid4().hex
                doc.id = doc_id

            self._add_to_index(doc, tokens)

    def search(self, query: Union[Query, str]) -> List[Document]:
        if isinstance(query, str):
            query = parse_query(query)

        doc_ids = self._eval_query(query)

        docs = [self.documents[d_id] for d_id in doc_ids]

        return docs

    def delete(self, docs: List[Document] = None, ids: List[str] = None) -> int:
        if docs is None and ids is None:
            raise Exception("docs or ids required to delete from index")

        ids_to_delete = []
        if docs:
            ids_to_delete = ids_to_delete + [
                d.id for d in docs if d.id in self.documents
            ]

        if ids:
            ids_to_delete = ids_to_delete + [id for id in ids if id in self.documents]

        # this doesn't seem very performant, could revisit to take advantage of bulk operations
        for d_id in ids_to_delete:
            doc = self.documents[d_id]

            # parses doc.text to tokens to clean up index, the tokens are not saved due to memory cost
            for tok in self.text_to_index_tokens(doc.text):
                if tok in self.inverted_index:
                    # TODO should handle better?
                    try:
                        self.inverted_index[tok].remove(d_id)
                    except ValueError:
                        pass
                    if len(self.inverted_index[tok]) == 0:
                        del self.inverted_index[tok]
                if tok in self.positional_index and d_id in self.positional_index[tok]:
                    del self.positional_index[tok][d_id]
                    if len(self.positional_index[tok]) == 0:
                        del self.positional_index[tok]

        # remove documents
        for d_id in ids_to_delete:
            del self.documents[d_id]

        return len(ids_to_delete)

    def save(self, path: str, mkdir: bool = True) -> bool:
        if not os.path.exists(path) and mkdir:
            Path(path).mkdir(parents=True, exist_ok=True)

        if not os.path.isdir(path):
            raise Exception(f"input path: {path} should be a folder")

        document_file_path = os.path.join(path, "docs.jsonl")
        index_file_path = os.path.join(path, "index.json")

        if os.path.exists(document_file_path):
            raise Exception(f"{document_file_path} already exists")
        if os.path.exists(index_file_path):
            raise Exception(f"{index_file_path} already exists")

        with open(document_file_path, "w") as doc_file:
            for d in self.documents.values():
                json.dump(d.model_dump(), doc_file)
                doc_file.write("\n")

        with open(index_file_path, "w") as index_file:
            file_body = {
                "token_normalizers": [
                    t.__class__.__name__ for t in self.token_normalizers
                ],
                "tokenizer": self.tokenizer.__class__.__name__,
                "inverted_index": self.inverted_index,
                "positional_index": self.positional_index,
            }
            json.dump(file_body, index_file)

        return True

    def load_from_file(self, path: str) -> bool:
        if not os.path.exists(path) or not os.path.isdir(path):
            raise Exception(f"{path} must be a existing folder")

        document_file_path = os.path.join(path, "docs.jsonl")
        index_file_path = os.path.join(path, "index.json")

        with open(index_file_path, "r") as f:
            # TODO may want to validate tokenizer + normalizer set up against file
            loaded_index = json.load(f)

        self.inverted_index = loaded_index["inverted_index"]
        self.positional_index = loaded_index["positional_index"]

        saved_docs = {}
        with open(document_file_path, "r") as f:
            for line in f:
                doc = Document.model_validate(json.loads(line))
                saved_docs[doc.id] = doc

        self.documents = saved_docs

        return True

    def _eval_query(self, query: Query) -> List[str]:
        if isinstance(query, BooleanQuery):
            and_set = None
            or_set = set()
            not_set = set()

            for clause in query.clauses:
                query = clause.query
                query_condition = clause.clause

                doc_ids = self._eval_query(query)

                if query_condition == Clause.MUST:
                    if and_set is None:
                        and_set = set(doc_ids)
                    else:
                        and_set = and_set.intersection(set(doc_ids))
                elif query_condition == Clause.SHOULD:
                    or_set.update(doc_ids)
                elif query_condition == Clause.MUST_NOT:
                    not_set.update(doc_ids)

            # if ANDs exists ORs are ignored
            match_doc_ids = and_set if and_set else or_set
            return match_doc_ids - not_set

        elif isinstance(query, TermQuery):
            # running same normalization on the search term to ensure consistency
            query_tokens = self._normalize_tokens([query.term])
            # TODO revisit: if normalization removes the token, consider no match
            if len(query_tokens) == 0:
                return []

            query_term = query_tokens[0]

            return self.inverted_index.get(query_term, [])

        elif isinstance(query, PhraseQuery):
            terms = self._normalize_tokens(query.terms)

            if len(terms) == 1:
                # if phrase query is normalized to 1 term, treat it like a TermQuery
                return self.inverted_index.get(query_term, [])
            elif len(terms) == 0:
                return []

            # +1 to mimic edit distance instead of word distance i.e. "word1 word2" should be edit distance of 0, but word distance of 1
            distance = query.distance + 1
            ordered = query.ordered

            postings = []
            for term in terms:
                if term not in self.positional_index:
                    return []
                postings.append(self.positional_index[term])

            doc_ids = []
            if len(terms) == 2:
                p1 = postings[0]
                p2 = postings[1]
                doc_ids = self._positional_intersect(p1, p2, distance, ordered)
            elif len(terms) > 2:
                doc_ids = self._multi_term_positional_intersect(
                    postings, distance, ordered
                )

            return doc_ids
        else:
            raise ValueError("Invalid Query type")

    def _find_match_doc_ids(self, search_index, to_search_index):
        match_doc_ids = []
        for doc_id in search_index.keys():
            if doc_id in to_search_index:
                match_doc_ids.append(doc_id)
        return match_doc_ids

    def _positional_intersect(self, p1: Dict, p2: Dict, k: int, ordered: bool):
        result = set()

        # iterate through the rarer term to find matching documents
        if len(p1.keys()) >= len(p2.keys()):
            doc_ids = self._find_match_doc_ids(p2, p1)
        else:
            doc_ids = self._find_match_doc_ids(p1, p2)

        for doc_id in doc_ids:
            temp = []
            positions1 = p1[doc_id]
            positions2 = p2[doc_id]

            for pp1 in positions1:
                for pp2 in positions2:
                    # if phrase search is order sensitive, skip when term two position is before term one
                    if ordered and pp2 < pp1:
                        continue

                    dis = abs(pp1 - pp2)
                    # != 0 checks the token is not on the same position i.e. "word word" would match doc="word"
                    if dis <= k and dis != 0:
                        temp.append(pp2)
                    elif pp2 > pp1:
                        break

                # this is here to allow saving of positions, should revisit in the future
                # potentially could reduce extra processing if we don't need the matched token index
                while len(temp) > 0 and abs(temp[0] - pp1) > k:
                    temp.remove(temp[0])

                for ps in temp:
                    # for now just return doc_id for simplicity
                    # result.append((doc_id, pp1, ps))
                    result.add(doc_id)

        # should revisit to clean up algo so maybe we don't need to construct set to list here
        # needed currently because matched doc_id can duplicate
        return list(result)

    def _multi_term_match_doc_ids(self, postings: List[Dict]):
        # start from the smallest candidate list to reduce search time
        sorted_postings = sorted(postings, key=lambda x: len(x.keys()))
        candidate = list(sorted_postings[0].keys())

        for posting in sorted_postings[1:]:
            candidate = [c for c in candidate if c in posting]

        return candidate

    def _multi_term_positional_intersect(
        self, postings: List[Dict], k: int, ordered: bool
    ):
        result_doc_ids = set()

        doc_ids = self._multi_term_match_doc_ids(postings)

        for doc_id in doc_ids:
            positions1 = postings[0][doc_id]
            positions2 = postings[1][doc_id]

            for pp1 in positions1:
                ranges = []
                # initialize search ranges, similar to two term phrase query
                for pp2 in positions2:
                    if ordered and pp2 < pp1:
                        continue

                    dis = abs(pp1 - pp2)
                    if dis <= k and dis != 0:
                        ranges.append((min(pp1, pp2), max(pp1, pp2)))
                    elif pp2 > pp1:
                        break

                for index, postings_k in enumerate(postings[2:]):
                    temp = []
                    positions_k = postings_k[doc_id]

                    for r in ranges:
                        for pp_k in positions_k:
                            if ordered and pp_k < r[1]:
                                continue

                            low = min(r[0], pp_k)
                            high = max(r[1], pp_k)
                            # - 1 - index subtracts the word distance count for matching tokens
                            # this way converts word distance into 'edit distance'
                            dis = high - low - 1 - index
                            if dis <= k and dis != 0:
                                temp.append((min(r[0], pp_k), max(r[1], pp_k)))
                            elif pp_k > r[1]:
                                break
                    ranges = temp

                if len(ranges) > 0:
                    result_doc_ids.add(doc_id)

        return list(result_doc_ids)



from textsearchpy.index import Index
from textsearchpy.query import Query
from typing import List
import time
import psutil


def create_index_from_data(data: List[str]):
    start = time.time()
    index = Index()
    index.append(data)
    end = time.time()
    elapsed_time = end - start
    print("Indexing Execution time:", elapsed_time, "seconds")
    print(f"Total Documents in Index: {len(index)}")
    print(f"Token Index Size: {len(index.inverted_index)}")

    return index


def print_memory_usage():
    process = psutil.Process()
    mem_usage = process.memory_info().rss / 1024**2
    print(f"Memory Usage: {mem_usage} MiB")


def evaluate_queries(index: Index, queries: List[Query]):
    runtimes = []

    print(f"---------evaluating {len(queries)} queries----------")

    for query in queries:
        start = time.time()
        docs = index.search(query)
        end = time.time()
        elapsed_time = end - start
        runtimes.append(elapsed_time)
        print(f"Found {len(docs)} Docs")

    print(f"Average query runtime: {sum(runtimes) / len(runtimes)}")
    print(f"Max query runtime: {max(runtimes)}")
    print(f"Min query runtime: {min(runtimes)}")
